# Jinja2 tools for cherrypy
# Copyright (C) 2021-2025 Patrik Dufresne
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import importlib
import logging
import time

import cherrypy
import jinja2

# Sentinel value
_UNDEFINED = object()

# Capture epoch time to invalidate cache of static file.
_cache_invalidate = int(time.time())


class Jinja2Tool(cherrypy.Tool):
    """
    Jinja2 Tool for CherryPy.
    """

    def __init__(self):
        super().__init__('before_handler', self._wrap_handler, 'jinja2', priority=30)

    def _finalize_assets(self, catalog, html):
        """
        Replace the placeholder token in the rendered HTML with the fully
        formatted asset tags, then reset asset state for the next render.
        """
        assets_html = catalog._format_collected_assets()
        assets_html = str(assets_html).replace('<script type="module" ', '<script ')
        catalog._emit_assets_later = False
        catalog.collected_css = []
        catalog.collected_js = []
        return str(html).replace(catalog._assets_placeholder, assets_html)

    def _wrap_handler(self, env, template, extra_processor=None, debug=False):

        def wrap(*args, **kwargs):
            # Call original handler
            context = self.oldhandler(*args, **kwargs)
            # Render template.
            return self.render_request(env=env, template=template, context=context, extra_processor=extra_processor)

        request = cherrypy.serving.request
        if request.handler is not None:
            # Replace request.handler with self
            if debug:
                cherrypy.log('replacing request handler', context='TOOLS.JINJA2', severity=logging.DEBUG)
            self.oldhandler = request.handler
            request.handler = wrap

    def create_env(self, package_name, filters={}, globals={}, tests={}):
        """
        Utility function used to create a default Jinja2 environment with good default.
        """
        env = jinja2.Environment(
            loader=jinja2.PackageLoader(package_name),
            auto_reload=True,
            autoescape=True,
            trim_blocks=True,
            lstrip_blocks=True,
        )
        # Param to bust caches
        env.globals['cache_invalidate'] = _cache_invalidate

        # Enable translation if available
        if hasattr(cherrypy.tools, 'i18n'):
            from .i18n import format_date, format_datetime, get_language_name, get_translation, ugettext, ungettext

            env.add_extension('jinja2.ext.i18n')
            env.install_gettext_callables(ugettext, ungettext, newstyle=True)
            env.filters['format_date'] = format_date
            env.filters['format_datetime'] = format_datetime
            env.globals['get_language_name'] = get_language_name
            env.globals['get_translation'] = get_translation

        # Update globals, filters and tests
        env.globals.update(globals)
        env.filters.update(filters)
        env.tests.update(tests)

        # Enable JinjaX if available
        try:
            import jinjax

            env.add_extension(jinjax.JinjaX)
            catalog = jinjax.Catalog(jinja_env=env, root_url="/static/components/")
            catalog.add_folder(importlib.resources.files(package_name) / 'components')
            catalog.add_folder(importlib.resources.files(__package__) / '..' / 'components')
        except ImportError:
            pass

        return env

    def render_request(self, template, context={}, env=_UNDEFINED, extra_processor=_UNDEFINED):
        """
        Render template for a given cherrypy request.
        """
        request = cherrypy.serving.request
        if env is _UNDEFINED:
            env = request.config.get('tools.jinja2.env')
        if extra_processor is _UNDEFINED:
            extra_processor = request.config.get('tools.jinja2.extra_processor')
        # Execute extra processor if defined.
        new_context = {}
        if extra_processor:
            new_context.update(extra_processor())
        new_context.update(context)
        # Render templates
        return self.render(env=env, template=template, context=new_context)

    def render(self, env, template, context={}):
        """
        Lower level function used to render a template using the given jinja2 environment, template(s) and variable context.
        """
        # Get the right templates
        if isinstance(template, (list, tuple)):
            names = [t.format(**context) for t in template]
            tmpl = env.select_template(names)
        else:
            tmpl = env.get_template(template)
        out = tmpl.render(context)

        # With JinjaX > 0.60 render explicitly here.
        if 'catalog' in env.globals and getattr(env.globals['catalog'], '_emit_assets_later', False):
            return self._finalize_assets(catalog=env.globals['catalog'], html=out)
        return out


cherrypy.tools.jinja2 = Jinja2Tool()



# Internationalisation tool for cherrypy
# Copyright (C) 2012-2025 Patrik Dufresne
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Internationalization and Localization for CherryPy

This tool provides locales and loads translations in the following order:

1. `with i18n.preferred_lang()`
2. `tools.i18n.func` callback function (optional)
3. HTTP Accept-Language headers
4. `tools.i18n.default` value

The tool uses `babel` <http://babel.edgewall.org> for localization and
handling translations. Within your Python code you can use four functions
defined in this module and the loaded locale provided as `i18n.get_translation()`

Example::

    from i18n import ugettext as _, ungettext

    class MyController(object):
        @cherrypy.expose
        def index(self):
            locale = cherrypy.response.i18n.locale
            s1 = _(u'Translateable string')
            s2 = ungettext(u'There is one string.',
                           u'There are more strings.', 2)
            return u'<br />'.join([s1, s2, locale.display_name])

If you have code (e.g. database models) that is executed before the response
object is available, use the *_lazy functions to mark the strings
translateable. They will be translated later on, when the text is used (and
hopefully the response object is available then).

Example::

    from i18n_tool import ugettext_lazy

    class Model:
        def __init__(self):
            name = ugettext_lazy(u'Name of the model')

For your templates read the documentation of your template engine how to
integrate babel with it. I think `Genshi<http://genshi.edgewall.org>`_ and
`Jinja 2<http://jinja.pocoo.org`_ support it out of the box.


Settings for the CherryPy configuration::

    [/]
    tools.i18n.on = True
    tools.i18n.default = Your language with territory (e.g. 'en_US')
    tools.i18n.mo_dir = Directory holding the locale directories
    tools.i18n.domain = Your gettext domain (e.g. application name)

The mo_dir must contain subdirectories named with the language prefix
for all translations, containing a LC_MESSAGES dir with the compiled
catalog file in it.

Example::

    [/]
    tools.i18n.on = True
    tools.i18n.default = 'en_US'
    tools.i18n.mo_dir = '/home/user/web/myapp/i18n'
    tools.i18n.domain = 'myapp'

    Now the tool will look for a file called myapp.mo in
    /home/user/web/myapp/i18n/en/LC_MESSACES/
    or generic: <mo_dir>/<language>/LC_MESSAGES/<domain>.mo

That's it.

:License: BSD
:Author: Thorsten Weimann <thorsten.weimann (at) gmx (dot) net>
:Date: 2010-02-08
"""


import os
import threading
from contextlib import contextmanager
from functools import lru_cache

import cherrypy
import pytz
from babel import dates
from babel.core import Locale, get_global
from babel.support import LazyProxy, NullTranslations, Translations

# Store current translation and preferred_lang
_current = threading.local()


def _get_config(key, default=None):
    """
    Lookup configuration from request, if available. Fallback to global config.
    """
    if getattr(cherrypy, 'request') and getattr(cherrypy.request, 'config') and key in cherrypy.request.config:
        return cherrypy.request.config[key]
    return cherrypy.config.get(key, default)


@contextmanager
def preferred_lang(lang):
    """
    Re-define the preferred language to be used for translation within a given context.

    with i18n.preferred_lang('fr'):
        i18n.gettext('some string')
    """
    assert lang is None or isinstance(lang, str)
    prev_lang = getattr(_current, 'preferred_lang', [])
    prev_trans = getattr(_current, 'translation', None)
    try:
        # Update prefered lang and clear translation.
        if lang:
            _current.preferred_lang = [lang]
        elif hasattr(_current, 'preferred_lang'):
            del _current.preferred_lang
        _current.translation = None
        yield
    finally:
        # Restore previous value
        _current.preferred_lang = prev_lang
        _current.translation = prev_trans


@contextmanager
def preferred_timezone(timezone):
    """
    Re-define the preferred timezone to be used for date format within a given context.

    with i18n.preferred_lang('America/Montreal'):
        i18n.format_datetime(...)
    """
    assert timezone is None or isinstance(timezone, str)
    prev_timezone = getattr(_current, 'preferred_timezone', [])
    prev_tzinfo = getattr(_current, 'tzinfo', None)
    try:
        # Update prefered lang and clear translation.
        _current.preferred_timezone = [timezone] + prev_timezone
        _current.tzinfo = None
        yield
    finally:
        # Restore previous value
        _current.preferred_timezone = prev_timezone
        _current.tzinfo = prev_tzinfo


@lru_cache(maxsize=10)
def _search_translation(dirname, domain, *locales):
    """
    Loads the first existing translations for known locale.

    :parameters:
        langs : List
            List of languages as returned by `parse_accept_language_header`.
        dirname : String
            Directory of the translations (`tools.I18nTool.mo_dir`).
            Might be a list of directories.
        domain : String
            Gettext domain of the catalog (`tools.I18nTool.domain`).

    :returns: Translations, the corresponding Locale object.
    """
    if not isinstance(locales, (list, tuple)):
        locales = [locales]
    t = Translations.load(dirname, locales, domain)
    # Ignore null translation
    if t.__class__ is NullTranslations:
        return None
    # Assign prefered local to this translation to know the current locale.
    trans_locale = Locale.parse(t.files[0].split('/')[-3])
    for locale in locales:
        try:
            locale = Locale.parse(locale)
            if trans_locale == locale or trans_locale.language == locale.language:
                t.locale = locale
                break
        except ValueError:
            continue  # Invalid locale identifier
    return t


def get_language_name(lang_code):
    """
    Translate the language code into it's language display name.
    """
    locale = Locale.parse(lang_code)
    trans = get_translation()
    return locale.get_language_name(trans.locale)


def get_timezone():
    """
    Get the best timezone information for the current context.
    """
    # When tzinfo is defined, use it
    tzinfo = getattr(_current, 'tzinfo', None)
    if tzinfo is not None:
        return tzinfo
    # Otherwise search for a valid timezone.
    tzinfo = None
    default_timezone = _get_config('tools.i18n.default_timezone')
    preferred_timezone = getattr(_current, 'preferred_timezone', [default_timezone])
    for timezone in preferred_timezone:
        try:
            tzinfo = dates.get_timezone(timezone)
            break
        except Exception:
            pass
    # If we can't find a valid timezone using the default and preferred value, fall back to server timezone.
    if tzinfo is None:
        tzinfo = dates.get_timezone(None)
    _current.tzinfo = tzinfo
    return _current.tzinfo


def get_translation():
    """
    Get the best translation for the current context.
    """
    # When translation is defined, use it
    translation = getattr(_current, 'translation', None)
    if translation is not None:
        return translation

    # Otherwise, we need to search the translation.
    # `preferred_lang` should always has a sane value within a cherrypy request because of hooks
    # But we also need to support calls outside cherrypy.
    default = _get_config('tools.i18n.default')
    preferred_lang = getattr(_current, 'preferred_lang', [default])
    mo_dir = _get_config('tools.i18n.mo_dir')
    domain = _get_config('tools.i18n.domain')
    trans = _search_translation(mo_dir, domain, *preferred_lang)
    if trans is None:
        trans = NullTranslations()
        trans.locale = Locale('en')
    _current.translation = trans
    return _current.translation


def list_available_locales():
    """
    Return a list of available translations.
    """
    mo_dir = _get_config('tools.i18n.mo_dir', False)
    domain = _get_config('tools.i18n.domain')
    if not mo_dir:
        return
    for lang in os.listdir(mo_dir):
        trans = _search_translation(mo_dir, domain, lang)
        if trans is not None:
            yield trans.locale


def list_available_timezones():
    """
    Return list of available timezone.
    """
    # Babel only support a narrow list of timezone.
    babel_timezone = get_global('zone_territories').keys()
    return [t for t in pytz.all_timezones if t in babel_timezone]


# Public translation functions
def ugettext(message):
    """Standard translation function. You can use it in all your exposed
    methods and everywhere where the response object is available.

    :parameters:
        message : Unicode
            The message to translate.

    :returns: The translated message.
    :rtype: Unicode
    """
    return get_translation().ugettext(message)


gettext = ugettext


def ungettext(singular, plural, num):
    """Like ugettext, but considers plural forms.

    :parameters:
        singular : Unicode
            The message to translate in singular form.
        plural : Unicode
            The message to translate in plural form.
        num : Integer
            Number to apply the plural formula on. If num is 1 or no
            translation is found, singular is returned.

    :returns: The translated message as singular or plural.
    :rtype: Unicode
    """
    return get_translation().ungettext(singular, plural, num)


ngettext = ungettext


def gettext_lazy(message):
    """Like ugettext, but lazy.

    :returns: A proxy for the translation object.
    :rtype: LazyProxy
    """

    def func():
        return get_translation().ugettext(message)

    return LazyProxy(func, enable_cache=False)


def format_datetime(datetime=None, format='medium', tzinfo=None):
    """
    Wraper arround babel format_datetime to provide a default locale.
    The timezone used to format the date is determine with the following priorities:
    * value of tzinfo
    * value of get_timezone()
    * default server time.
    """
    # When formating date or english, let use en_GB as it's to most complete translation.
    return dates.format_datetime(
        datetime=datetime,
        format=format,
        locale=get_translation().locale,
        tzinfo=tzinfo or get_timezone(),
    )


def format_date(datetime=None, format='medium', tzinfo=None):
    """
    Wraper arround babel format_date to provide a default locale.
    """
    return format_datetime(datetime=datetime, format=dates.get_date_format(format), tzinfo=tzinfo)


def get_timezone_name(tzinfo):
    locale = Locale(get_translation().locale.language)
    return dates.get_timezone_name(tzinfo, width='long', locale=locale)


def _load_default(mo_dir, domain, default, **kwargs):
    """
    Initialize the language using the default value from the configuration.
    """
    # Clear current translation
    _current.preferred_lang = [default]
    _current.preferred_timezone = [kwargs['default_timezone']] if 'default_timezone' in kwargs else []
    cherrypy.request._i18n_lang_func = kwargs.get('lang', kwargs.get('func', False))
    cherrypy.request._i18n_tzinfo_func = kwargs.get('tzinfo', False)
    # Clear current translation
    _current.translation = None
    # Clear current timezone
    _current.tzinfo = None


def _load_accept_language(**kwargs):
    """
    When running within a request, load the prefered language from Accept-Language
    """

    if cherrypy.request.headers.elements('Accept-Language'):
        count = 0
        for x in cherrypy.request.headers.elements('Accept-Language'):
            _current.preferred_lang.insert(count, x.value.replace('-', '_'))
            count += 1
    # Clear current translation
    _current.translation = None


def _load_func_language(**kwargs):
    """
    When running a request where a current user is found, load prefered language from user preferences.
    """
    func = getattr(cherrypy.request, '_i18n_lang_func', False)
    if not func:
        return
    try:
        lang = func()
    except Exception:
        return
    if not lang:
        return
    # Add custom lang to preferred_lang
    _current.preferred_lang.insert(0, lang)
    # Clear current translation
    _current.translation = None


def _load_func_tzinfo(**kwargs):
    """
    When running a request, load the prefered timezone information from user preferences.
    """
    func = getattr(cherrypy.request, '_i18n_tzinfo_func', False)
    if not func:
        return
    try:
        tzinfo = func()
    except Exception:
        return
    if not tzinfo:
        return
    # Add custom lang to preferred_lang
    _current.preferred_timezone.insert(0, tzinfo)
    # Clear current translation
    _current.tzinfo = None


def _set_content_language(**kwargs):
    """
    Sets the Content-Language response header (if not already set) to the
    language of `cherrypy.response.i18n.locale`.
    """
    if 'Content-Language' not in cherrypy.response.headers:
        locale = get_translation().locale
        language_tag = f"{locale.language}-{locale.territory}" if locale.territory else locale.language
        cherrypy.response.headers['Content-Language'] = language_tag


class I18nTool(cherrypy.Tool):
    """Tool to integrate babel translations in CherryPy."""

    def __init__(self):
        super().__init__('before_handler', _load_default, 'i18n')

    def _setup(self):
        cherrypy.Tool._setup(self)
        # Attach additional hooks as different priority to update preferred lang with more accurate preferences.
        cherrypy.request.hooks.attach('before_handler', _load_accept_language, priority=60)
        cherrypy.request.hooks.attach('before_handler', _load_func_language, priority=75)
        cherrypy.request.hooks.attach('before_handler', _load_func_tzinfo, priority=75)
        cherrypy.request.hooks.attach('before_finalize', _set_content_language)


cherrypy.tools.i18n = I18nTool()



# CherryPy Foundation
# Copyright (C) 2025 IKUS Software inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from urllib.parse import urljoin

import cherrypy


def url_for(*args, _relative=None, _base=None, **kwargs):
    """
    Generate a URL for the given endpoint/path (*args) with query params (**kwargs).

    relative:
      - None: pass through to cherrypy.url (CherryPy default behavior).
      - False: absolute URL (scheme, host, vhost, script_name).
      - True: URL relative to current request path (may include '..').
      - 'server': URL relative to server root (starts with '/').
    Notes:
      - String chunks have leading/trailing slashes stripped and are joined with '/'.
      - Chunks '.' and '..' are allowed as literals (for relative paths).
      - Objects may implement __url_for__() -> str or have a string .url_for attribute.
      - Integers are appended as path segments.
      - When path == "", existing request query parameters are merged (kwargs win).
    """
    # Handle query-string
    qs = [(k, v) for k, v in sorted(kwargs.items()) if v is not None]

    path = []
    for chunk in args:
        if hasattr(chunk, '__url_for__') and callable(chunk.__url_for__):
            path.append(str(chunk.__url_for__()))
        elif hasattr(chunk, 'url_for'):
            path.append(str(chunk.url_for))
        elif isinstance(chunk, str):
            path.append(chunk)
        elif isinstance(chunk, int):
            path.append(str(chunk))
        else:
            raise ValueError('invalid positional arguments, url_for accept str, bytes, int: %r' % chunk)
    path = '/'.join(path)
    # When path is empty, we are browsing the same page.
    # Let keep the original query_string to avoid loosing it.
    if not path:
        params = cherrypy.request.params.copy()
        params.update(kwargs)
        qs = [(k, v) for k, v in sorted(params.items()) if v is not None]
    elif not path.startswith('.'):
        path = urljoin('/', path)
    # Outside a request, use cherrypy.tools.proxy config
    if not cherrypy.request.app:
        if _base is None:
            _base = cherrypy.config.get('tools.proxy.base', None)
    # Use cherrypy to build the URL
    return cherrypy.url(path=path, qs=qs, relative=_relative, base=_base)



# CherryPy Foundation
# Copyright (C) 2020-2025 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import namedtuple

import cherrypy

FlashMessage = namedtuple('FlashMessage', ['message', 'level'])


def flash(message, level='info'):
    """
    Add a flash message to the session.
    """
    assert message
    assert level in ['info', 'error', 'warning', 'success']
    if 'flash' not in cherrypy.session:
        cherrypy.session['flash'] = []
    # Support Markup and string
    if hasattr(message, '__html__'):
        flash_message = FlashMessage(message, level)
    else:
        flash_message = FlashMessage(str(message), level)
    cherrypy.session['flash'].append(flash_message)


def get_flashed_messages():
    """
    Return all flash message.
    """
    if 'flash' in cherrypy.session:
        messages = cherrypy.session['flash']
        del cherrypy.session['flash']
        return messages
    return []



# CherryPy Foundation
# Copyright (C) 2025 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import importlib
from unittest import skipUnless

import cherrypy
from cherrypy.test import helper

import cherrypy_foundation.tools.jinja2  # noqa
from cherrypy_foundation.flash import flash, get_flashed_messages

HAS_JINJAX = importlib.util.find_spec("jinjax") is not None

env = cherrypy.tools.jinja2.create_env(
    package_name=__package__,
    globals={'get_flashed_messages': get_flashed_messages},
)


@cherrypy.tools.sessions()
@cherrypy.tools.jinja2(env=env)
class Root:

    @cherrypy.expose
    @cherrypy.tools.jinja2(template='test_flash.html')
    def index(self):
        flash('default flash message', level='info')
        return {}


@skipUnless(HAS_JINJAX, reason='Required jinjax')
class FormTest(helper.CPWebCase):
    default_lang = None
    interactive = False

    @classmethod
    def setup_server(cls):
        cherrypy.tree.mount(Root(), '/')

    def test_get_flash(self):
        # Given a page returning a flash message
        # When querying the page that include this form
        self.getPage("/")
        self.assertStatus(200)
        # Then page display the message.
        self.assertInBody('test-flash')
        self.assertInBody('<div class="alert alert-info alert-dismissible fade show"')
        self.assertInBody('default flash message')



# Cherrypy Foundation
# Copyright (C) 2025 IKUS Software inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import importlib.resources

import cherrypy

RE_STATIC_MATCH = r"(\.js|\.css|\.png|\.woff|\.woff2|\.map)$"


@cherrypy.config(
    **{
        'tools.auth.on': False,
        'tools.auth_mfa.on': False,
        'tools.i18n.on': False,
        'tools.ratelimit.on': False,
        'tools.secure_headers.on': False,
        'tools.sessions.on': False,
    }
)
class StaticMiddleware:

    @cherrypy.expose
    @cherrypy.tools.staticdir(match=RE_STATIC_MATCH, dir=str(importlib.resources.files(__package__)))
    def default(self, *args, **kwargs):
        """This entry point is used to serve content of /static/components/ folder and JinjaX static ressources"""
        # Make use of JinjaX catalog
        env = cherrypy.request.config.get('tools.jinja2.env')
        if env is None or 'catalog' not in env.globals:
            raise cherrypy.HTTPError(400)

        # JinjaX resources could be locaed in multiple path.
        section = cherrypy.serving.request.toolmaps['tools']['staticdir']['section']
        jinjax_catalog = env.globals['catalog']
        for path in jinjax_catalog.paths:
            handled = cherrypy.lib.static.staticdir(section=section, match=RE_STATIC_MATCH, dir=path)
            if handled:
                return cherrypy.serving.response.body
        raise cherrypy.HTTPError(400)



# CherryPy Foundation
# Copyright (C) 2025 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import importlib
from datetime import datetime, timezone
from unittest import skipUnless

import cherrypy
from cherrypy.test import helper
from parameterized import parameterized

from .. import i18n  # noqa
from .. import jinja2  # noqa

HAS_JINJAX = importlib.util.find_spec("jinjax") is not None

env = cherrypy.tools.jinja2.create_env(
    package_name=__package__,
    globals={
        'const1': 'STATIC VALUE',
    },
)


def extra_processor():
    return {'var2': 'bar'}


@cherrypy.tools.jinja2(env=env, extra_processor=extra_processor)
class Root:

    @cherrypy.expose
    @cherrypy.tools.jinja2(template='test_jinja2.html')
    def index(self):
        return {'var1': 'test-jinja2'}

    @cherrypy.expose
    @cherrypy.tools.jinja2(template='test_jinjax.html')
    def jinjax(self):
        return {'var1': 'test-jinjax'}

    @cherrypy.expose
    @cherrypy.tools.jinja2(template='test_jinja2_i18n.html')
    @cherrypy.tools.i18n(
        default='fr',
        default_timezone='America/Toronto',
        mo_dir=importlib.resources.files(__package__) / 'locales',
        domain='messages',
    )
    def localized(self):
        return {
            'my_datetime': datetime(year=2025, month=11, day=26, hour=11, minute=16, tzinfo=timezone.utc),
            'my_date': datetime(year=2025, month=12, day=22, hour=14, minute=8, tzinfo=timezone.utc),
        }


class Jinja2Test(helper.CPWebCase):
    default_lang = None
    interactive = False

    @classmethod
    def setup_server(cls):
        cherrypy.tree.mount(Root(), '/')

    def test_get_page(self):
        # Given a page render using jinja2
        # When querying the page
        self.getPage("/")
        # Then the page return without error
        self.assertStatus(200)
        # Then the page is render dynamically using page context
        self.assertInBody('test-jinja2')
        self.assertInBody('bar')
        self.assertInBody('STATIC VALUE')

    @skipUnless(HAS_JINJAX, reason='Required jinjax')
    def test_get_page_jinjax(self):
        # Given a page render using jinjax
        # When querying the page
        self.getPage("/jinjax")
        # Then the page return without error
        self.assertStatus(200)
        # Then the page is render dynamically using page context
        self.assertInBody('<a class="btn btn-primary" href="http://example.com">foo</a>')

    @parameterized.expand(
        [
            ('server_default', {}, 'fr'),
            ('accept_lang_fr', {'Accept-Language': 'fr'}, 'fr'),
            ('accept_lang_en', {'Accept-Language': 'en'}, 'en'),
        ]
    )
    def test_get_page_i18n(self, _name, headers, expected_lang):
        # Given a localized page render with jinja2
        # When querying the page
        self.getPage("/localized", headers=list(headers.items()))
        # Then the page return without error
        self.assertStatus(200)
        # Then the page is render dynamically using page context
        if expected_lang == 'fr':
            self.assertInBody('lang="fr"')
            self.assertInBody('français')
            self.assertInBody('Du texte à traduire')
            self.assertInBody('mercredi 26 novembre 2025, 06:16:00 heure normale de l’Est nord-américain')
            self.assertInBody('lundi, décembre 22, 2025')
        else:
            self.assertInBody('lang="en"')
            self.assertInBody('English')
            self.assertInBody('Some text to translate')
            self.assertInBody('Wednesday, November 26, 2025, 6:16:00\u202fAM Eastern Standard Time')
            self.assertInBody('Monday, December 22, 2025')



# Session timeout tool for cherrypy
# Copyright (C) 2012-2025 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import math
import time

import cherrypy
from cherrypy.lib import httputil

SESSION_PERSISTENT = '_session_persistent'
SESSION_START_TIME = '_session_start_time'

SESSION_DEFAULT_ABSOLUTE_TIMEOUT = 43200  # 30 days
SESSION_DEFAULT_PERSISTENT_TIMEOUT = 10080  # 7 days
SESSION_DEFAULT_TIMEOUT = 60  # 60 minutes (from cherrypy default)


class SessionsTimeout(cherrypy.Tool):
    """
    Fine-grained control over session timeouts.

    Config keys:
      - tools.sessions.timeout (int, minutes)  -> base idle timeout (CherryPy built-in, used for non-persistent)
      - tools.sessions_timeout.absolute_timeout (int, minutes)  -> hard cap for all sessions (default=43200 = 30d)
      - tools.sessions_timeout.persistent_timeout (int, minutes) -> sliding window for persistent sessions (default=10080 = 7d)
      - tools.sessions.name (str) -> session cookie name (default: 'session_id')
    """

    def __init__(self, priority: int = 75):
        super().__init__(point='before_handler', callable=self.run, priority=priority)

    @staticmethod
    def _cookie_name() -> str:
        return cherrypy.request.config.get('tools.sessions.name', 'session_id')

    @staticmethod
    def _get_config_timeouts() -> tuple[int, int, int]:
        cfg = cherrypy.request.config
        idle_timeout = int(cfg.get('tools.sessions.timeout', SESSION_DEFAULT_TIMEOUT))
        persistent_timeout = int(
            cfg.get('tools.sessions_timeout.persistent_timeout', SESSION_DEFAULT_PERSISTENT_TIMEOUT)
        )
        absolute_timeout = int(cfg.get('tools.sessions_timeout.absolute_timeout', SESSION_DEFAULT_ABSOLUTE_TIMEOUT))
        return idle_timeout, persistent_timeout, absolute_timeout

    def _set_cookie_max_age(self, max_age: int) -> None:
        """Adjust only Max-Age and Expires for the session cookie, keep other flags intact."""
        cookie_name = self._cookie_name()
        cookie = cherrypy.serving.response.cookie
        # Ensure the cookie key exists (CherryPy sets it when session id is issued/regenerated).
        if cookie_name not in cookie:
            cookie[cookie_name] = cherrypy.serving.session.id
        cookie[cookie_name]['max-age'] = str(max(0, max_age))
        cookie[cookie_name]['expires'] = httputil.HTTPDate(time.time() + max(0, max_age))

    @staticmethod
    def _ceil_minutes(seconds: int) -> int:
        if seconds <= 0:
            return 0
        return int(math.ceil(seconds / 60.0))

    def _ensure_start_time(self) -> None:
        """Ensure the session has a stable start time anchor for absolute timeout."""
        session = cherrypy.serving.session
        if SESSION_START_TIME not in session or not isinstance(session.get(SESSION_START_TIME), datetime.datetime):
            session[SESSION_START_TIME] = session.now()

    def _expire_and_restart(self, make_persistent: bool | None = None) -> None:
        """Expire current session and start a fresh one. Optionally set persistent flag."""
        session = cherrypy.serving.session
        session.clear()
        session.regenerate()
        session[SESSION_START_TIME] = session.now()
        if make_persistent is not None:
            session[SESSION_PERSISTENT] = bool(make_persistent)

    def _update_session_timeout(self) -> None:
        session = cherrypy.serving.session
        self._ensure_start_time()

        now: datetime.datetime = session.now()
        start: datetime.datetime = session[SESSION_START_TIME]

        idle_timeout_min, persistent_timeout_min, absolute_timeout_min = self._get_config_timeouts()

        is_persistent = bool(session.get(SESSION_PERSISTENT, False))

        # Compute absolute remaining (hard cap from start, never slides)
        abs_expiration = start + datetime.timedelta(minutes=absolute_timeout_min)
        abs_remaining_sec = int((abs_expiration - now).total_seconds())

        if is_persistent:
            # Sliding window for persistent sessions
            sliding_exp = now + datetime.timedelta(minutes=persistent_timeout_min)
        else:
            # Sliding window for non-persistent sessions (idle timeout)
            sliding_exp = now + datetime.timedelta(minutes=idle_timeout_min)

        sliding_remaining_sec = int((sliding_exp - now).total_seconds())

        # Effective remaining is the minimum of the sliding window and absolute cap
        remaining_sec = min(abs_remaining_sec, sliding_remaining_sec)
        remaining_min = self._ceil_minutes(remaining_sec)

        if remaining_min <= 0:
            # Expired due to either sliding window or absolute cap
            # Start a fresh session; do NOT automatically re-mark persistent here.
            # Authentication/Remember me logic elsewhere can call set_persistent(True) again after login.
            self._expire_and_restart()
            # After regeneration, set a sensible session.timeout baseline:
            cherrypy.serving.session.timeout = idle_timeout_min
            return

        # Apply per-request remaining minutes to CherryPy's session timeout
        session.timeout = remaining_min

        if is_persistent:
            # Keep cookie lifetime aligned with remaining time so it can survive browser restarts.
            self._set_cookie_max_age(remaining_sec)
        else:
            # For non-persistent, let CherryPy manage the session cookie (session or transient cookie).
            # If you want strict enforcement client-side too, uncomment the next line:
            # self._set_cookie_max_age(remaining_sec)
            pass

    def run(self, persistent_timeout: int = 10080, absolute_timeout: int = 43200) -> None:
        # Skip if sessions are not enabled
        if not hasattr(cherrypy.serving, 'session'):
            return

        # Ensure configured values are honored (method signature kept for CherryPy Tool interface)
        self._update_session_timeout()

    def set_persistent(self, value: bool = True) -> None:
        """Mark current session as persistent (or not) and reset the start window."""
        session = cherrypy.serving.session
        session[SESSION_PERSISTENT] = bool(value)
        # Reset absolute window anchor when the persistence policy changes
        session[SESSION_START_TIME] = session.now()
        self._update_session_timeout()

    def is_persistent(self) -> bool:
        """Return True if the session is marked persistent."""
        return bool(cherrypy.serving.session.get(SESSION_PERSISTENT, False))

    def get_session_start_time(self) -> datetime.datetime | None:
        """Return the session start time if any."""
        return cherrypy.serving.session.get(SESSION_START_TIME, None)


cherrypy.tools.sessions_timeout = SessionsTimeout()



# CherryPy Foundation
# Copyright (C) 2025 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import importlib
from unittest import skipUnless

import cherrypy
from cherrypy.test import helper
from wtforms.fields import BooleanField, PasswordField, StringField, SubmitField
from wtforms.validators import InputRequired, Length

import cherrypy_foundation.tools.jinja2  # noqa
from cherrypy_foundation.form import CherryForm
from cherrypy_foundation.tools.i18n import gettext_lazy as _

HAS_JINJAX = importlib.util.find_spec("jinjax") is not None

env = cherrypy.tools.jinja2.create_env(
    package_name=__package__,
)


class LoginForm(CherryForm):
    login = StringField(
        _('User'),
        validators=[
            InputRequired(),
            Length(max=256, message=_('User too long.')),
        ],
        render_kw={
            "placeholder": _('User'),
            "autocorrect": "off",
            "autocapitalize": "none",
            "autocomplete": "off",
            "autofocus": "autofocus",
        },
    )
    password = PasswordField(
        _('Password'),
        validators=[
            InputRequired(),
            Length(max=256, message=_('Password too long.')),
        ],
        render_kw={"placeholder": _("Password")},
    )
    persistent = BooleanField(
        _('Remember me'),
        # All `label-*` are assigned to the label tag.
        render_kw={'container_class': 'col-sm-6', 'label-attr': 'FOO'},
    )
    submit = SubmitField(
        _('Login'),
        # All `container-*` are assigned to the container tag.
        render_kw={"class": "btn-primary float-end", 'container_class': 'col-sm-6', 'container-attr': 'BAR'},
    )


@cherrypy.tools.jinja2(env=env)
class Root:

    @cherrypy.expose
    @cherrypy.tools.jinja2(template='test_form.html')
    def index(self):
        form = LoginForm()
        return {'form': form}


@skipUnless(HAS_JINJAX, reason='Required jinjax')
class FormTest(helper.CPWebCase):
    default_lang = None
    interactive = False

    @classmethod
    def setup_server(cls):
        cherrypy.tree.mount(Root(), '/')

    def test_get_form(self):
        # Given a form
        # When querying the page that include this form
        self.getPage("/")
        self.assertStatus(200)
        # Then each field is render properly.
        # 1. Check title
        self.assertInBody('test-form')
        # 2. Check user field
        self.assertInBody('<label class="form-label" for="login">User</label>')
        self.assertInBody(
            '<input autocapitalize="none" autocomplete="off" autocorrect="off" autofocus="autofocus" class="form-control" id="login" maxlength="256" name="login" placeholder="User" required type="text" value="">'
        )
        # 3. Check password
        self.assertInBody('<label class="form-label" for="password">Password</label>')
        self.assertInBody(
            '<input class="form-control" id="password" maxlength="256" name="password" placeholder="Password" required type="password" value="">'
        )
        # 4 Check remember me
        self.assertInBody(
            '<input class="form-check-input" container-class="col-sm-6" id="persistent" label-attr="FOO" name="persistent" type="checkbox" value="y">'
        )
        self.assertInBody('<label attr="FOO" class="form-check-label" for="persistent">Remember me</label>')
        # 5. check submit button (regex matches because class could have different order with jinjax<=0.57)
        self.assertInBody('<div  attr="BAR"')
        self.assertMatchesBody(
            '<input class="(btn-primary ?|float-end ?|btn ?){3}" container-attr="BAR" container-class="col-sm-6" id="submit" name="submit" type="submit" value="Login">'
        )



# Scheduler plugins for Cherrypy
# Copyright (C) 2025 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime, timezone
from threading import Event

import cherrypy
from cherrypy.test import helper

from .. import scheduler  # noqa

done = Event()


def a_task(*args, **kwargs):
    done.set()


class SchedulerPluginTest(helper.CPWebCase):
    def setUp(self) -> None:
        done.clear()
        cherrypy.scheduler.remove_all_jobs()
        return super().setUp()

    def tearDown(self):
        return super().tearDown()

    @classmethod
    def setup_server(cls):
        pass

    def test_add_job(self):
        # Given a scheduled job
        scheduled = cherrypy.engine.publish(
            'scheduler:add_job',
            a_task,
            name='custom_name',
            args=(1, 2, 3),
            kwargs={'foo': 'bar'},
            next_run_time=datetime.now(timezone.utc),
            misfire_grace_time=None,
        )
        self.assertTrue(scheduled)
        self.assertEqual('custom_name', scheduled[0].name)
        # When waiting for all jobs
        cherrypy.scheduler.wait_for_jobs()
        # Then the job is done
        self.assertTrue(done.is_set())

    def test_add_job_daily(self):
        # Given a scheduler with a specific number of jobs
        count = len(cherrypy.scheduler.get_jobs())
        # When scheduling a daily job.
        scheduled = cherrypy.engine.publish('scheduler:add_job_daily', '23:00', a_task, 1, 2, 3, foo=1, bar=2)
        # Then the job is scheduled
        self.assertTrue(scheduled)
        self.assertEqual('a_task', scheduled[0].name)
        # Then the number of jobs increase.
        self.assertEqual(count + 1, len(cherrypy.scheduler.get_jobs()))

    def test_add_job_now(self):
        # Given a task
        # When scheduling that task
        scheduled = cherrypy.engine.publish('scheduler:add_job_now', a_task, 1, 2, 3, foo=1, bar=2)
        self.assertTrue(scheduled)
        # When waiting for all tasks
        cherrypy.scheduler.wait_for_jobs()
        # Then the task get called
        self.assertTrue(done.is_set())

    def test_remove_job(self):
        # Given a scheduler with a specific number of jobs
        count = len(cherrypy.scheduler.get_jobs())
        # Given a job schedule every seconds
        cherrypy.engine.publish('scheduler:add_job_daily', '23:00', a_task, 1, 2, 3, foo=1, bar=2)
        # Then number of job increase
        self.assertEqual(count + 1, len(cherrypy.scheduler.get_jobs()))
        # When the job is unscheduled.
        cherrypy.engine.publish('scheduler:remove_job', a_task)
        # Then the number of job decrease.
        self.assertEqual(count, len(cherrypy.scheduler.get_jobs()))

    def test_remove_job_with_invalid_job(self):
        # Given an unschedule job
        # When unscheduling an invalid job
        cherrypy.engine.publish('scheduler:remove_job', a_task)
        # Then an error is not raised.



# Cherrypy Foundation
# Copyright (C) 2022-2025 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import importlib
import tempfile
import unittest
from urllib.parse import urlencode

import cherrypy
from cherrypy.test import helper

HAS_SQLALCHEMY = importlib.util.find_spec("sqlalchemy") is not None

if not HAS_SQLALCHEMY:
    pass
else:
    from sqlalchemy import Column, Integer, String, func
    from sqlalchemy.exc import IntegrityError
    from sqlalchemy.sql.schema import Index

    from .. import db  # noqa

    Base = cherrypy.db.get_base()

    class User(Base):
        __tablename__ = 'users'
        id = Column(Integer, primary_key=True)
        username = Column(String)

        def __repr__(self):
            return f"User(id={self.id}, username='{self.username}')"

    Index('user_username_unique_ix', func.lower(User.username), unique=True, info='This username already exists.')

    class Root:

        @cherrypy.expose
        def index(self):
            return str(User.query.all())

        @cherrypy.expose
        def add(self, username):
            try:
                User(username=username).add().commit()
                return "OK"
            except IntegrityError as e:
                return str(e)

    @unittest.skipUnless(HAS_SQLALCHEMY, "sqlalchemy not installed")
    class DbPluginTest(helper.CPWebCase):
        interactive = False

        @classmethod
        def setup_class(cls):
            cls.tempdir = tempfile.TemporaryDirectory(prefix='cherrypy-foundation-', suffix='-db-test')
            super().setup_class()

        @classmethod
        def teardown_class(cls):
            cls.tempdir.cleanup()
            super().teardown_class()

        @classmethod
        def setup_server(cls):
            cherrypy.config.update(
                {
                    'db.uri': f"sqlite:///{cls.tempdir.name}/data.db",
                }
            )
            cherrypy.tree.mount(Root(), '/')

        def setUp(self):
            cherrypy.db.create_all()
            return super().setUp()

        def tearDown(self):
            cherrypy.db.drop_all()
            return super().tearDown()

        def test_add_user(self):
            # Given an empty database
            self.assertEqual(0, User.query.count())
            # When calling "/add/" to create a user
            self.getPage('/add', method='POST', body=urlencode({'username': 'myuser'}))
            self.assertStatus(200)
            self.assertInBody('OK')
            # Then user get added to database
            self.assertEqual(1, User.query.count())

        def test_add_duplicate_user(self):
            # Given a database with a users
            User(username='user1').add().commit()
            # When trying to add another user
            self.getPage('/add', method='POST', body=urlencode({'username': 'user1'}))
            # Then an error get raised
            self.assertStatus(200)
            self.assertInBody('user_username_unique_ix')

        def test_get_users(self):
            # Given a database with a users
            new_user = User(username='newuser').add().commit()
            # When calling "/"
            self.getPage("/")
            self.assertStatus(200)
            # Then the page include our users
            self.assertInBody(new_user.username)



# CherryPy
# Copyright (C) 2025 IKUS Software
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import cherrypy
from cherrypy.test import helper
from parameterized import parameterized

from ..error_page import error_page


class Root:

    @cherrypy.expose
    def index(self):
        return "OK"

    @cherrypy.expose
    def not_found(self):
        raise cherrypy.NotFound()

    @cherrypy.expose
    def not_found_custom(self):
        raise cherrypy.HTTPError(404, message='My error message')

    @cherrypy.expose
    def html_error(self):
        raise cherrypy.HTTPError(400, message='My error message')

    @cherrypy.expose
    @cherrypy.tools.json_out()
    def json_error(self):
        raise cherrypy.HTTPError(400, message='json error message')

    @cherrypy.expose
    @cherrypy.tools.response_headers(headers=[('Content-Type', 'text/plain')])
    def text_error(self):
        raise cherrypy.HTTPError(400, message='text error message')


class ErrorPageTest(helper.CPWebCase):
    interactive = False

    @classmethod
    def setup_server(cls):
        cherrypy.config.update(
            {
                'error_page.default': error_page,
            }
        )
        cherrypy.tree.mount(Root(), '/')

    @parameterized.expand(
        [
            ('/not_found', '<p>Nothing matches the given URI</p>'),
            ('/not_found_custom', '<p>My error message</p>'),
            ('/html_error', '<p>My error message</p>'),
            ('/json_error', '{"message": "json error message", "status": "400 Bad Request"}'),
            ('/text_error', 'text error message'),
        ]
    )
    def test_error_page(self, page, expect_body):
        # When query return an error
        self.getPage(page)
        # then error page adjust the content
        self.assertInBody(expect_body)
